"""
Use the GitHub API to show the status of current GitHub Action runs on a
specific branch.

It tries to show enough runs to see the latest status of each workflow.  Runs
older than a week are not considered.

"""

import contextlib
import io
import os
import re
import signal
import sys
import urllib.parse

from os.path import isdir

import click
import exceptiongroup
import rich.console

from .data_core import draw_runs
from .git_help import git_repo_urls, git_branch
from .http_help import get_data
from .utils import Interval, WatchGhaError


console = rich.console.Console(highlight=False)
error_console = rich.console.Console(stderr=True, highlight=False)


def fatal(msg, status=1):
    error_console.print(msg)
    sys.exit(status)


@contextlib.contextmanager
def handle_resize(handler):
    original_sigwinch_handler = None
    try:
        # Save the original handler.
        original_sigwinch_handler = signal.getsignal(signal.SIGWINCH)
        # Call the given handler on sigwinch.
        signal.signal(signal.SIGWINCH, lambda _, __: handler())
    except AttributeError:
        # This system is probably windows, and doesn't have SIGWINCH.
        # Unfortunately there's no signal for window resize on windows and I
        # don't know how to properly handle it. Just swallow the error.
        pass

    # yield control to the with block, and re-raise any exceptions
    try:
        yield
    except:
        raise
    finally:
        # clean up our sigwinch handler if we set one
        if original_sigwinch_handler:
            signal.signal(signal.SIGWINCH, original_sigwinch_handler)


@click.command()
@click.option("--sha", help="The commit SHA to use. Must be a full SHA.")
@click.option(
    "--poll",
    help="How many seconds between refreshes.",
    type=int,
    default=15,
    show_default=True,
)
@click.option("--wait-for-start", is_flag=True, help="Wait for jobs to start.")
@click.option(
    "--only",
    help=(
        "Words to limit the workflows shown. "
        + "Only workflows with these comma separated case insensitive substrings "
        + "in their names will be shown."
    ),
)
@click.argument("repo", default=".")
@click.argument("branch", required=False)
def main(sha, poll, wait_for_start, only, repo, branch):
    """
    Watch GitHub Action runs.

    Repeatedly gets the latest status and redraws the screen, until all of the
    jobs are complete.

    REPO is a local directory or GitHub URL, defaulting to ".".

    BRANCH is defaulted from the git repo.

    """
    if only is not None:
        only_words = [w.strip() for w in only.split(",")]
    else:
        only_words = None

    watcher = GhaWatcher(
        urls=gha_urls(repo, branch, sha),
        get_data_fn=get_data,
        only_words=only_words,
    )

    watcher.watch(wait_for_start, poll, console)


def gha_urls(repo, branch=None, sha=None):
    """Figure out the GHA api URLs to use for `repo`, `branch`, and `sha`."""
    if isdir(repo):
        repo_urls = list(git_repo_urls(repo))
        if branch is None:
            branch = git_branch(repo)
    elif ":" in repo:
        repo_urls = [repo]
        if branch is None:
            fatal(f"Branch is required for URL repo")
    else:
        fatal(f"Don't understand repo {repo!r}")

    params = {"per_page": "100"}
    if sha:
        params["head_sha"] = sha
    else:
        assert branch is not None
        params["branch"] = branch
    url_args = urllib.parse.urlencode(params)

    github_urls = []
    for repo_url in repo_urls:
        # repo_url = "https://github.com/owner/repo.git"
        # repo_url = "git@github.com:someorg/somerepo.git"
        # see also https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
        server_url = os.getenv("GITHUB_SERVER_URL", "https://github.com")
        repo_match = re.fullmatch(
            rf"(?:{re.escape(server_url)}|git@github.com)[/:]?([^/]+/[^/]+?)(?:\.git|/)?",
            repo_url,
        )
        if repo_match is None:
            continue

        api_url = os.getenv("GITHUB_API_URL", "https://api.github.com")
        url = f"{api_url}/repos/{repo_match[1]}/actions/runs?{url_args}"
        github_urls.append(url)

    if not github_urls:
        fatal(f"Couldn't find GitHub repo from remote URLs: {repo_urls!r}")

    return github_urls


class GhaWatcher:
    def __init__(self, urls, get_data_fn, only_words):
        self.urls = urls
        self.get_data_fn = get_data_fn
        self.only_words = only_words
        self.status = 0
        self.error = None

    def watch(self, wait_for_start, poll, console):
        self.done = False
        self.succeeded = False
        self.interrupted = False

        self.watch_gha_errors = []
        interval = Interval(poll)

        output = ""

        with exceptiongroup.catch(
            {
                WatchGhaError: self.handle_watchghaerror,
                KeyboardInterrupt: self.handle_keyboardinterrupt,
            }
        ):
            while True:
                output = self.get_gha_display()
                if wait_for_start:
                    if not self.done:
                        break
                else:
                    break

            if not self.done:
                with console.screen() as screen:
                    with handle_resize(lambda: screen.update(output)):
                        while not self.done:
                            screen.update(output)
                            interval.wait()
                            output = self.get_gha_display()

        if self.watch_gha_errors:
            fatal(self.watch_gha_errors[0])
        console.print(output, end="")
        if self.interrupted:
            fatal("** interrupted **", status=2)
        sys.exit(0 if self.succeeded else 1)

    def handle_watchghaerror(self, excgroup):
        self.watch_gha_errors.extend(excgroup.exceptions)

    def handle_keyboardinterrupt(self, excgroup):
        self.interrupted = True

    def get_gha_display(self):
        stream = io.StringIO()

        self.done, self.succeeded = draw_runs(
            self.urls,
            datafn=self.get_data_fn,
            outfn=lambda s: print(s, file=stream),
            only_words=self.only_words,
        )
        return stream.getvalue()



import pytest

from watchgha import watch_runs
from watchgha.watch_runs import gha_urls


def fake_isdir(path):
    return path == "." or path.startswith("dir")


FAKE_REPOS = {
    ".": {
        "urls": [
            "https://github.com/owner/repo.git",
        ],
        "branch": "mainbranch",
    },
    "dir1": {
        "urls": [
            "git@github.com:somebody/therepo.git",
        ],
        "branch": "joe/feature1",
    },
    "dir-many-remotes": {
        "urls": [
            "https://gitlab.com/maintainer/project.git",
            "https://github.com/contributor1/project",
            "https://github.com/contributor2/project.git",
        ],
        "branch": "main",
    },
    "dir-only-gitlab": {
        "urls": [
            "https://gitlab.com/maintainer/project.git",
        ],
    },
    "dir-enterprise": {
        "urls": [
            "https://mygithub.enterprise.com/me/myproject",
        ],
    },
    "dir-bug22": {
        "urls": [
            "git@git.mydomain.com:davidszotten/davidrepo",
        ],
    },
}


def fake_git_repo_urls(repo):
    return FAKE_REPOS.get(repo, {}).get("urls", ())


def fake_git_branch(repo):
    return FAKE_REPOS.get(repo, {}).get("branch", "master")


@pytest.fixture
def mocked_gha_urls_dependencies(monkeypatch):
    monkeypatch.setattr(watch_runs, "isdir", fake_isdir)
    monkeypatch.setattr(watch_runs, "git_repo_urls", fake_git_repo_urls)
    monkeypatch.setattr(watch_runs, "git_branch", fake_git_branch)


@pytest.mark.parametrize(
    "args, evars, urls",
    [
        (
            ["."],
            {},
            [
                "https://api.github.com/repos/owner/repo/actions/runs?per_page=100&branch=mainbranch",
            ],
        ),
        (
            ["dir1"],
            {},
            [
                "https://api.github.com/repos/somebody/therepo/actions/runs?per_page=100&branch=joe%2Ffeature1",
            ],
        ),
        (
            ["dir1", "another-branch"],
            {},
            [
                "https://api.github.com/repos/somebody/therepo/actions/runs?per_page=100&branch=another-branch",
            ],
        ),
        (
            ["dir1", None, "sha12345678"],
            {},
            [
                "https://api.github.com/repos/somebody/therepo/actions/runs?per_page=100&head_sha=sha12345678",
            ],
        ),
        (
            ["dir-many-remotes"],
            {},
            [
                "https://api.github.com/repos/contributor1/project/actions/runs?per_page=100&branch=main",
                "https://api.github.com/repos/contributor2/project/actions/runs?per_page=100&branch=main",
            ],
        ),
        (
            ["."],
            {"GITHUB_API_URL": "https://theapi.nedhub.com"},
            [
                "https://theapi.nedhub.com/repos/owner/repo/actions/runs?per_page=100&branch=mainbranch",
            ],
        ),
        (
            ["https://github.com/me/myproject.git", "mybranch"],
            {},
            [
                "https://api.github.com/repos/me/myproject/actions/runs?per_page=100&branch=mybranch",
            ],
        ),
        (
            ["dir-enterprise"],
            {
                "GITHUB_SERVER_URL": "https://mygithub.enterprise.com",
                "GITHUB_API_URL": "https://api.mygithub.enterprise.com",
            },
            [
                "https://api.mygithub.enterprise.com/repos/me/myproject/actions/runs?per_page=100&branch=master",
            ],
        ),
        (
            ["dir-bug22"],
            {
                "GITHUB_SERVER_URL": "git@git.mydomain.com",
                "GITHUB_API_URL": "https://githubapi.mydomain.com",
            },
            [
                "https://githubapi.mydomain.com/repos/davidszotten/davidrepo/actions/runs?per_page=100&branch=master",
            ],
        ),
    ],
)
def test_gha_urls(monkeypatch, mocked_gha_urls_dependencies, args, evars, urls):
    for name, value in evars.items():
        monkeypatch.setenv(name, value)
    assert gha_urls(*args) == urls


@pytest.mark.parametrize(
    "args, evars, msg",
    [
        (
            ["https://mygithub.com/me/myproject.git"],
            {},
            "Branch is required for URL repo",
        ),
        (
            ["what!?"],
            {},
            "Don't understand repo 'what!?'",
        ),
        (
            ["dir-only-gitlab"],
            {},
            "Couldn't find GitHub repo from remote URLs: \n['https://gitlab.com/maintainer/project.git']",
        ),
    ],
)
def test_gha_urls_fails(
    capsys, monkeypatch, mocked_gha_urls_dependencies, args, evars, msg
):
    with pytest.raises(SystemExit):
        gha_urls(*args)
    assert capsys.readouterr().err.strip() == msg



import datetime
import re
import time


class WatchGhaError(Exception):
    pass


def nice_time(dt):
    dt = dt.astimezone()
    now = datetime.datetime.now()
    if dt.date() != now.date():
        if dt.year != now.year:
            fmt = "%Y-%m-%d %I:%M%p"
        else:
            fmt = "%m-%d %I:%M%p"
    else:
        fmt = "%I:%M%p"
    return dt.strftime(fmt).lower()


def to_datetime(isostr):
    # 3.11 accepts Z, but older Pythons don't.
    isostr = isostr.replace("Z", "+00:00")
    return datetime.datetime.fromisoformat(isostr)


class DictAttr:
    def __init__(self, d):
        self.d = d

    def __getattr__(self, name):
        return self.d[name]


class Interval:
    """Wait for an interval of time to pass.

    Better than time.sleep because it accounts for time spent doing things
    other than sleeping.

    """

    def __init__(self, secs):
        self.secs = secs
        self.last_time = time.time()

    def wait(self):
        now = time.time()
        delay = self.secs - (now - self.last_time)
        self.last_time = now
        if delay > 0:
            time.sleep(delay)


def human_key(s):
    """Turn a string into a sortable value that works how humans expect.

    "z23A" -> (["z", 23, "a"], "z23A")

    The original string is appended as a last value to ensure the
    key is unique enough so that "x1y" and "x001y" can be distinguished.
    """
    def tryint(s: str) -> str | int:
        """If `s` is a number, return an int, else `s` unchanged."""
        try:
            return int(s)
        except ValueError:
            return s

    return ([tryint(c) for c in re.split(r"(\d+)", s.casefold())], s)
